; *****************************************************************************
;
; "21 or Bust" in Blitz
;
; regrettably, by eelhovercraft / Evil Roy Ferguson
; Coded 2006-ish I think?
;
; I wrote this several years ago for my dad, and it was sloppy even for then.
; I only spent an hour or two on it and it shows pretty hard.
; I do not recommend many of the coding practices demonstrated
; in this game.
;
; I found the code today while I was looking for something entirely
; unrelated, but when I Googled "21 or Bust," I wasn't able to find
; any versions of the game that were not affiliated with one or more
; casinos of the online and sleazy persuasion.
;
; I'm really not sure what's going on with my hungarian notation here.
; Pretend it's not there. :(
;
; *****************************************************************************

; Screen dimensions
Const cScreenWidth = 640
Const cScreenHeight = 480

; Size of a deck
Const cDeckSize = 52+4

; The number of columns and the maximum number of cards they can hold
Const cColumnCount = 4
Const cColumnSize = 4

; Width of a pile
Const cColumnSpacing = 80

; Where the piles appear on the screen
Const cLeftOffset = (cScreenWidth - cColumnSpacing*(cColumnCount+1))/2

; Points for a match and for clearing the deck
Const cMatchScore = 10
Const cClearScore = 50

; which frame is the card back? 
Const cCardBack = 52

; messages
Const cGameName$ 		= "21 or Bust"
Const cGameOverMsg$ 	= "Game Over!"
Const cDeckClearedMsg$  = "Deck Cleared!"
Const cScoreText$		= "Score: "

; The names of the suits and cards. Never actually used now.
Dim aSuitNames$(3)
Dim aCardNames$(14)

; The cards in each column, and the combined total of the cards in it
Dim aColumns.TCard(cColumnCount, cColumnSize)
Dim aColumnVal(cColumnCount)

; Our card type
Type TCard
	Field suit
	Field value
	Field frame		; frame in the cards AnimImage
End Type

; The deck, the discard pile, and slots.
Dim aDeck.TCard(cDeckSize)
Dim aDiscard.TCard(cDeckSize)
Dim aSlots(cDeckSize)

; The current height of the deck and the discard pile.
Global gDiscardHeight, gDeckHeight

; images
Global gCards
Global gHighlight

; The player's current score
Global gScore

; Used to allow undo
Global gLastPile = -1
Global gLastSlot = -1
Global gLastVal = 0

; Currently highlighted column
Global gHighlighted = -1

Global gMessage$ = "", gMessageSet = False, gMessageOn = False
Global gGameOver = False

; blragh
Main()

; set all cards in all piles to null
Function ClearPiles()

	Local i, j ; set everybuddy to null
	For i = 0 To cColumnCount
		For j = 0 To cColumnSize
			aColumns(i,j) = Null
		Next
		aColumnVal(i) = 0
	Next
	
End Function

; move all cards to the discard pile
Function DiscardAll()

	Local card.TCard
	
	gDiscardHeight = 0
	gDeckHeight = 0
	
	For card = Each TCard
		aDiscard(gDiscardHeight) = card
		gDiscardHeight = gDiscardHeight + 1
	Next
	
	ClearPiles()

End Function

; Creates one of each type of card.
Function CreateCards()
	; clear existing cards, we're through with them
	Delete Each TCard
		
	; make sure one of each card is created
	Local is, iv
	Local card.TCard
	
	For is = 0 To 3 ; loop through suits
		For iv = 1 To 13 ; loop through values
		
			card = New TCard ; create a card
			card\suit = is ; assign appropriate suit and value
			card\value = iv
			
			; assign the correct frame to the card.
			; This is necessary because the image is unusual.
			card\frame = 13 * card\suit
			If card\value > 1
				card\frame = card\frame + 14-card\value
			EndIf
			
			; deal with jokers, although they aren't used in 21 Pick Up
			If card\value = 0
				card\frame = 53 + (card\suit>1)
			EndIf
						
		Next ; values
	Next ; suits
	
End Function

; Shuffles the deck and is not exactly space-efficient :(
Function ShuffleDeck()

	; initialize "slots"
	Local i
	For i = 0 To gDiscardHeight-1
		aSlots(i) = i
	Next
	
	Local slot
	Local index	
	
	gDeckHeight = gDiscardHeight
		
	; loop while unplaced cards exist
	While gDiscardHeight > 0
	
		; decrement the size of the discard stack
		gDiscardHeight = gDiscardHeight - 1
	
		; find an unoccupied "slot"
		slot = Rand(0, gDiscardHeight)		
		
		; this is the index within the deck that slot points to
		index = aSlots(slot)
				
		; put the first unplaced card in the deck at that slot		
		aDeck(index) = aDiscard(gDiscardHeight)
		
		; Copy the slot furthest in the list to the current cell
		; (which is now occupied and shouldn't be used again)
		; then reduce the number of remaining slots by one.
		aSlots(slot) = aSlots(gDiscardHeight)		
	
	Wend
	
	; the deck is now shuffled. Long live Jambi.

End Function

; Loads the name of each card
Function InitNames()

	Local iSuit
	
	Restore SuitNames
	For i = 0 To 3
		Read aSuitNames(i)
	Next
	
	Restore CardNames
	For i = 0 To 13
		Read aCardNames(i)
	Next

End Function

; Returns the name of a card
; InitNames() must have been called
Function CardName$(card.TCard)
	Return aCardNames(card\value) + " of " + aSuitNames(card\suit)
End Function

; Removes the top card from the deck and returns it 
Function DrawCard.TCard()
	
	If gDeckHeight =< 0 Return Null
	
	gDeckHeight = gDeckHeight - 1
	Return aDeck(gDeckHeight)
	
End Function

; Returns the top card on the pile without removing it
Function PeekCard.TCard()
	
	If gDeckHeight =< 0 Return Null
	
	Return aDeck(gDeckHeight - 1)
	
End Function

; Discards a card
Function DiscardCard(card.TCard)
	
	aDiscard(gDiscardHeight) = card
	gDiscardHeight = gDiscardHeight + 1
	
End Function

; Gets the value of a card for a particular pile
; Aces are worth either 1 or 11 depending on what you "need"
Function CardVal(card.TCard, pile)

	; get a card's "real" value, varies on pile
	If card\value < 11
		; special handling of the ace
		If card\value = 1
			If aColumnVal(pile) > 10
				Return 1
			Else
				Return 11
			EndIf
		EndIf
		Return card\value
	Else				
		Return 10
	EndIf
	
End Function

; Sets the message
Function SetMessage(msg$)
	gMessage = msg$
	gMessageSet = True
	gMessageOn = True
End Function

; Plays a card onto a pile
Function PlayCard(pile, card.TCard)

	Local i
	Local val = CardVal(card, pile)
	Local colval = aColumnVal(pile)
					
	; stack space?
	For i = 0 To cColumnSize
		If aColumns(pile, i) = Null
			Exit
		EndIf
	Next		
	
	; yes, if card value is small enough, it's playable
	If colval + val =< 21
		; space left, put it on the stack
		If i =< cColumnSize
			aColumns(pile, i) = DrawCard()			
			
			gLastPile = pile
			gLastSlot = i
			gLastVal = aColumnVal(pile)
			
			aColumnVal(pile) = colval + val
		EndIf																		

		; cleared!
		If aColumnVal(pile) = 21 Or i > cColumnSize
			; scoreage!
			DumpPile(pile)
				
			gScore = gScore + cMatchScore
			
			gLastPile = -1
			gLastSlot = -1
		EndIf
		
		; if deck is empty shuffle it and give the player points
		If gDeckHeight < 1		
			ShuffleDeck()
			gScore = gScore + cClearScore
			SetMessage(cDeckClearedMsg)
		EndIf
		
	EndIf
	
End Function

; discards every card in a pile and sets its value to 0
Function DumpPile(pile)

	Local i
	
	For i = 0 To cColumnSize
		If aColumns(pile, i) <> Null
			DiscardCard(aColumns(pile, i))
			aColumns(pile, i) = Null
		EndIf
	Next
	
	aColumnVal(pile) = 0
	
End Function

; Previews the deck
Function DrawPreview(x, y)

	Local previews = gDeckHeight / 4
	Local i	
	
	; draw "deck height"
	For i = 1 To previews
		DrawImage gCards, x, y, cCardBack
		x = x + 2
	Next
	
	; draw top card of deck
	If gDeckHeight > 0
		DrawImage gCards, x, y, aDeck(gDeckHeight-1)\frame
	EndIf

End Function

; Draws every pile
Function DrawPiles()

	Local i, j

	; draw the heavenly light of the chosen one
	If gHighlighted > -1
		DrawBlock gHighlight, cLeftOffset - 4 + gHighlighted * 80, 160
	EndIf

	For i = 0 To cColumnCount ; okay, go ahead and make my day
		OutText cLeftOffset + i * cColumnSpacing + cColumnSpacing/2, 135, aColumnVal(i), 1
		For j = 0 To cColumnSize ; wryyyyyy
			If aColumns(i,j) <> Null
				DrawImage gCards, cLeftOffset + i * cColumnSpacing, 165 + j * 25, aColumns(i,j)\frame					
			EndIf
		Next
	Next
	
End Function

; Draws the scoreboard
Function DrawScore()
	OutText 400, 30, cScoreText
	OutText 540, 30, gScore, 1
	
	If gLastPile > -1
		OutText 400, 80, "Right-click to undo."
	EndIf
End Function

; Starts a new game
Function ResetGame()

	DiscardAll()	
	ShuffleDeck()
	
	gScore = 0
	gLastPile = -1
	gLastSlot = -1
	gLastVal = 0
	
	gGameOver = False

End Function

Function PlayerInput()

	Local cx = (MouseX()-cLeftOffset) / 80
	Local redraw = 0
	
	If MouseY() > 150
		If cx > -1 And cx =< cColumnCount	
			;TODO separate logic from drawing!
			If cx <> gHighlighted			
				gHighlighted = cx
			EndIf
			
			If MouseHit(1)
				PlayCard(cx, aDeck(gDeckHeight-1))
			EndIf
		EndIf
	EndIf
	
	If MouseHit(2) ; best undo implementation evar.
		If gLastPile > -1		
			Local card.TCard = aColumns(gLastPile, gLastSlot)
					
			aDeck(gDeckHeight) = card
			gDeckHeight = gDeckHeight + 1
			
			aColumnVal(gLastPile) = gLastVal
			aColumns(gLastPile, gLastSlot) = Null
			
			gLastPile = -1
			gLastSlot = -1			
		EndIf
	EndIf
	
End Function

; Are we fail?
Function CheckLoss()

	Local card.TCard = PeekCard()	
	Local i
	Local fails
	
	For i = 0 To cColumnCount
		If CardVal(card, i) + aColumnVal(i) > 21
			fails = fails + 1
		EndIf
	Next
	
	If fails = cColumnCount+1
		Return 1 ; fail
	Else
		Return 0 ; hm
	EndIf

End Function

; Load the cards and things
Function LoadResources()

	gCards = LoadAnimImage("carddeck.png", 68, 96, 0, 55)	
	MaskImage gCards, 255, 0, 255
		
	gHighlight = LoadImage("highlight.png")
	
	SetFont LoadFont("Verdana", 25, 1)
	ClsColor 0, 128, 0

End Function

; shows text with a lovely outline
Function OutText(x, y, s$, cx=False, cy=False)

	Color 0, 0, 0
	Text x - 1, y - 1, s, cx, cy
	Text x + 1, y - 1, s, cx, cy
	Text x - 1, y, s, cx, cy
	Text x + 1, y, s, cx, cy
	Text x - 1, y + 1, s, cx, cy
	Text x + 1, y + 1, s, cx, cy

	Color 255, 255, 255
	Text x, y, s, cx, cy

End Function

; Does what it says on the tin
Function RedrawGame()

	; a more efficient implementation would redraw only the parts of the game that have changed
	; this is not a more efficient implementation. I am lazy.
	Cls
	
	DrawPreview(20, 20)		
	DrawScore()
	DrawPiles()
	
	OutText GraphicsWidth()/2, GraphicsHeight()/2, gMessage, True, True	
	
	Flip False

End Function

; Ohnoes
Function Main()

	; Seeed are the masters of german reggae, but I prefer to pretend that
	; the genre simply doesn't exist
	SeedRnd MilliSecs()

	; go go gadget graphics.
	AppTitle cGameName
	Graphics cScreenWidth, cScreenHeight, 0, 2
	SetBuffer BackBuffer()
	
	; load some cards
	LoadResources()
	
	; initialize our lovely suits, no longer called since... well, no longer used...
	;InitNames()
	
	; create our fabulous cards
	CreateCards()
	
	; fabulous!
	ResetGame()
	
	; get an initial screen goin'
	RedrawGame()
	
	; Whee
	Local timer = CreateTimer(15)
	
	; Main loop
	Repeat			
		
		; You lose, good day sir
		If CheckLoss()			
			SetMessage(cGameOverMsg)
			gGameOver = True
		EndIf
		
		; If there's a message, show it.
		If gMessageOn	
			If MouseHit(1)
				; Nummy nummy
				gMessageOn = False
				If gGameOver
					ResetGame()					
				EndIf
				gMessage = ""
			EndIf
		Else
			; hubba hubba zoot zoot				
			PlayerInput()	
		EndIf
						
		; deba uba zat zat, a num num
		RedrawGame()
		
		; a-hoorepa hoorepa a-huh-hoorepa a-num num
		Delay 2
	
	Until KeyDown(1)
	
End Function

.SuitNames
; I originally had "cloves" here. I'm not sure why.
Data "Clubs", "Spades", "Diamonds", "Hearts"

.CardNames
Data "Joker", "Ace", "Two", "Three", "Four", "Five", "Six", "Seven"
Data "Eight", "Nine", "Ten", "Jack", "Queen", "King"